from django.template import Template, Context
from django.conf import settings
from django.test import TestCase, override_settings
from shorturls.tests.models import Animal, Vegetable, Mineral


@override_settings(
    ROOT_URLCONF='shorturls.urls',
    SHORTEN_MODELS={
        'A': 'shorturls.animal',
        'V': 'shorturls.vegetable',
    },
)
class TemplateTagTestCase(TestCase):
    fixtures = ['shorturls-test-data.json']

    def setUp(self):
        settings.SHORT_BASE_URL = None

    def render(self, t, **c):
        return Template('{% load shorturl %}' + t).render(Context(c))

    def test_shorturl(self):
        r = self.render('{% shorturl a %}', a=Animal.objects.get(id=12345))
        self.assertEqual(r, '/ADNH')

    def test_bad_context(self):
        r = self.render('{% shorturl a %}')
        self.assertEqual(r, '')

    def test_no_prefix(self):
        r = self.render('{% shorturl m %}', m=Mineral.objects.all()[0])
        self.assertEqual(r, '')

    def test_short_base_url(self):
        settings.SHORT_BASE_URL = 'http://example.com/'
        r = self.render('{% shorturl a %}', a=Animal.objects.get(id=12345))
        self.assertEqual(r, 'http://example.com/ADNH')

    def test_revcanonical(self):
        r = self.render('{% revcanonical a %}', a=Animal.objects.get(id=12345))
        self.assertEqual(r, '<link rev="canonical" href="/ADNH">')



"""
A handful of test modules to test out resolving redirects.
"""

from django.db import models


class Animal(models.Model):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = 'shorturls'

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return '/animal/{0!s}/'.format(self.id)


class Vegetable(models.Model):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = 'shorturls'

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return 'http://example.net/veggies/{0!s}'.format(self.id)


class Mineral(models.Model):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = 'shorturls'

    def __unicode__(self):
        return self.name



from django.conf import settings
from django.http import HttpResponse
from django.urls import re_path

from . import views


def handler404(req, exception=None):
    return HttpResponse(status=404)

urlpatterns = [
    re_path(
        r'^(?P<prefix>{0!s})(?P<tiny>\w+)$'.format(
            '|'.join(settings.SHORTEN_MODELS.keys())),
        views.redirect,
    ),
]



from django.conf import settings
from django.test import TestCase, override_settings
from shorturls.baseconv import base62


@override_settings(
    ROOT_URLCONF='shorturls.urls',
    SHORTEN_MODELS={
        'A': 'shorturls.animal',
        'V': 'shorturls.vegetable',
        'M': 'shorturls.mineral',
        'bad': 'not.amodel',
        'bad2': 'not.even.valid',
    },
    SHORTEN_FULL_BASE_URL='http://example.com',
)
class RedirectViewTestCase(TestCase):
    fixtures = ['shorturls-test-data.json']

    def test_redirect(self):
        """
        Test the basic operation of a working redirect.
        """
        response = self.client.get('/A{0!s}'.format(enc(12345)))
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'],
                         'http://example.com/animal/12345/')

    def test_redirect_from_request(self):
        """
        Test a relative redirect when the Sites app isn't installed.
        """
        settings.SHORTEN_FULL_BASE_URL = None
        response = self.client.get(
            '/A{0!s}'.format(enc(54321)), HTTP_HOST='example.org')
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'],
                         'http://example.org/animal/54321/')

    def test_redirect_complete_url(self):
        """
        Test a redirect when the object returns a complete URL.
        """
        response = self.client.get('/V{0!s}'.format(enc(785)))
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'],
                         'http://example.net/veggies/785')

    def test_bad_short_urls(self):
        self.assertEqual(404, self.client.get('/badabcd').status_code)
        self.assertEqual(404, self.client.get('/bad2abcd').status_code)
        self.assertEqual(404, self.client.get('/Vssssss').status_code)

    def test_model_without_get_absolute_url(self):
        self.assertEqual(404, self.client.get(
            '/M{0!s}'.format(enc(10101))).status_code)


def enc(id):
    return base62.from_decimal(id)



DEBUG = True
SECRET_KEY = '123'
ALLOWED_HOSTS = ['*']

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3'
    }
}

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'shorturls',
]
ROOT_URLCONF = 'shorturls.urls'

SHORTEN_MODELS = {
    'A': 'shorturls.animal',
    'V': 'shorturls.vegetable',
    'M': 'shorturls.mineral',
    'bad': 'not.amodel',
    'bad2': 'not.even.valid',
}
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
    },
]



from urllib.parse import urljoin, urlsplit

from django.apps import apps
from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.http import HttpResponsePermanentRedirect, Http404
from django.shortcuts import get_object_or_404
from shorturls import default_converter


def redirect(request, prefix, tiny, converter=default_converter):
    """
    Redirect to a given object from a short URL.
    """
    # Resolve the prefix and encoded ID into a model object and decoded ID.
    # Many things here could go wrong -- bad prefix, bad value in
    # SHORTEN_MODELS, no such model, bad encoding -- so just return a 404 if
    # any of that stuff goes wrong.
    try:
        app_label, model_name = settings.SHORTEN_MODELS[prefix].split('.')
    except (KeyError, ValueError):
        raise Http404('Bad prefix.')
    try:
        model = apps.get_model(app_label, model_name)
    except LookupError:
        model = False
    if not model:
        raise Http404('Bad model specified in SHORTEN_MODELS.')
    try:
        id = converter.to_decimal(tiny)
    except ValueError:
        raise Http404('Bad encoded ID.')

    # Try to look up the object. If it's not a valid object, or if it doesn't
    # have an absolute url, bail again.
    obj = get_object_or_404(model, pk=id)
    try:
        url = obj.get_absolute_url()
    except AttributeError:
        raise Http404(
            "'{0!s}' models don't have a get_absolute_url() method.".format(model.__name__))

    # We might have to translate the URL -- the badly-named get_absolute_url
    # actually returns a domain-relative URL -- into a fully qualified one.

    # If we got a fully-qualified URL, sweet.
    if urlsplit(url)[0]:
        return HttpResponsePermanentRedirect(url)

    # Otherwise, we need to make a full URL by prepending a base URL.
    # First, look for an explicit setting.
    if hasattr(settings, 'SHORTEN_FULL_BASE_URL') and settings.SHORTEN_FULL_BASE_URL:
        base = settings.SHORTEN_FULL_BASE_URL

    # Finally, fall back on the current request or site.
    else:
        base = 'http://{0!s}/'.format(get_current_site(request).domain)

    return HttpResponsePermanentRedirect(urljoin(base, url))



from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from shorturls import baseconv

default_converter = baseconv.base62

if hasattr(settings, 'SHORTURLS_DEFAULT_CONVERTER'):
    mod_name, conv_name = settings.SHORTURLS_DEFAULT_CONVERTER.rsplit('.', 1)
    try:
        mod = import_module(mod_name)
    except ImportError as e:
        raise ImproperlyConfigured(
            'Could not load converter specified by SHORTURLS_DEFAULT_CONVERTER. Error was: {0!s}'.format(e))
    try:
        default_converter = getattr(mod, conv_name)
    except AttributeError:
        raise ImproperlyConfigured(
            'Could not load converter specified by SHORTURLS_DEFAULT_CONVERTER. {0!s} is not in {1!s}.'.format(conv_name, mod))



from urllib.parse import urljoin

from django import template
from django.conf import settings
from django import urls as urlresolvers
from django.utils.safestring import mark_safe
from shorturls import default_converter as converter, views


class ShortURL(template.Node):

    @classmethod
    def parse(cls, parser, token):
        parts = token.split_contents()
        if len(parts) != 2:
            raise template.TemplateSyntaxError(
                "{0!s} takes exactly one argument".format(parts[0]))
        return cls(template.Variable(parts[1]))

    def __init__(self, obj):
        self.obj = obj

    def render(self, context):
        try:
            obj = self.obj.resolve(context)
        except template.VariableDoesNotExist:
            return ''

        try:
            prefix = self.get_prefix(obj)
        except (AttributeError, KeyError):
            return ''

        tinyid = converter.from_decimal(obj.pk)

        if hasattr(settings, 'SHORT_BASE_URL') and settings.SHORT_BASE_URL:
            return urljoin(settings.SHORT_BASE_URL, prefix + tinyid)

        try:
            return urlresolvers.reverse(views.redirect, kwargs={
                'prefix': prefix,
                'tiny': tinyid
            })
        except urlresolvers.NoReverseMatch:
            return ''

    def get_prefix(self, model):
        if not hasattr(self.__class__, '_prefixmap'):
            self.__class__._prefixmap = dict(
                (m, p) for p, m in settings.SHORTEN_MODELS.items())
        key = '{0!s}.{1!s}'.format(
            model._meta.app_label, model.__class__.__name__.lower())
        return self.__class__._prefixmap[key]


class RevCanonical(ShortURL):

    def render(self, context):
        url = super(RevCanonical, self).render(context)
        if url:
            return mark_safe('<link rev="canonical" href="{0!s}">'.format(url))
        else:
            return ''

register = template.Library()
register.tag('shorturl', ShortURL.parse)
register.tag('revcanonical', RevCanonical.parse)



import unittest
from shorturls import baseconv


class BaseConvTests(unittest.TestCase):

    def _test_converter(self, converter):
        nums = [-10 ** 10, 10 ** 10] + list(range(-100, 100))
        for before in nums:
            after = converter.to_decimal(converter.from_decimal(before))
            self.assertEqual(before, after)

    def test_bin(self):
        self._test_converter(baseconv.bin)

    def test_hex(self):
        self._test_converter(baseconv.hexconv)

    def test_base62(self):
        self._test_converter(baseconv.base62)

    def test_base32(self):
        self._test_converter(baseconv.base32)

    def test_base32_mappings(self):
        converter = baseconv.base32
        self.assertEqual(
            converter.from_decimal(converter.to_decimal('Fo9')),
            'f09'
        )
        self.assertEqual(
            converter.from_decimal(converter.to_decimal('43FG5i2L')),
            '43fg5121'
        )



"""
Convert numbers from base 10 integers to base X strings and back again.

Original: http://www.djangosnippets.org/snippets/1431/

Sample usage:

>>> base20 = BaseConverter('0123456789abcdefghij')
>>> base20.from_decimal(1234)
'31e'
>>> base20.to_decimal('31e')
1234
"""


class BaseConverter(object):
    decode_mapping = {}

    def __init__(self, digits):
        self.digits = digits
        self.length = len(digits)

    def from_decimal(self, i):
        if i < 0:
            i, neg = -i, 1
        else:
            neg = 0
        enc = ''
        while i >= self.length:
            i, mod = divmod(i, self.length)
            enc = self.digits[mod] + enc
        enc = self.digits[i] + enc
        if neg:
            enc = '-' + enc
        return enc

    def to_decimal(self, s):
        if self.decode_mapping:
            new = ''
            for digit in s:
                if digit in self.decode_mapping:
                    new += self.decode_mapping[digit]
                else:
                    new += digit
            s = new
        if str(s)[0] == '-':
            s, neg = str(s)[1:], 1
        else:
            neg = 0
        decoded = 0
        multi = 1
        while len(s) > 0:
            decoded += multi * self.digits.index(s[-1:])
            multi = multi * self.length
            s = s[:-1]
        if neg:
            decoded = -decoded
        return decoded

bin = BaseConverter('01')
hexconv = BaseConverter('0123456789ABCDEF')
base62 = BaseConverter(
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789abcdefghijklmnopqrstuvwxyz'
)


class Base32Converter(BaseConverter):
    """
    http://www.crockford.com/wrmg/base32.html
    """
    decode_mapping = {
        'o': '0',
        'i': '1',
        'l': '1',
    }

    def __init__(self):
        super(Base32Converter, self).__init__(
            '0123456789abcdefghjkmnpqrstvwxyz')

    def to_decimal(self, s):
        return super(Base32Converter, self).to_decimal(s.lower())

base32 = Base32Converter()

if __name__ == '__main__':
    import doctest
    doctest.testmod()
